/***** pim.c **************************************
 * Description: Compute the estimator of the number
 *   of pairwise mismatches from shustring lengths
 *   as described in Haubold, Reed, Pfaffelhuber
 *   (2010). Alignment-free estimation of 
 *   nucleotide diversity identifies a
 *   novel genomic outlier of unusually low genetic
 *   diversity in Drosophila melanogaster. In
 *   preparation.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * Date: Thu Mar 12 15:40:15 2009
 **************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "eprintf.h"
#include "sequenceData.h"
#include "interface.h"
#include "lcpTree.h"
#include "pim.h"

void windowAnalysis(int *sl, int n, int w, int stepLen);

int main(int argc, char *argv[]){
  int queryDscr;
  Args *args;
  char *version;
  int i;

  version = "0.3";
  setprogname2("pim");
  args = getArgs(argc, argv);
  if(args->p)
    printSplash(version);
  if(args->h || args->e)
    printUsage(version);
  if(args->numInputFiles == 0){
    queryDscr = 0;
    runAnalysis(queryDscr, args);
  }else{
    for(i=0;i<args->numInputFiles;i++){
      queryDscr = open(args->inputFiles[i],0);
      if(queryDscr < 0)
	eprintf("ERROR: could not open query file %s\n",args->inputFiles[i]);
      runAnalysis(queryDscr, args);
      close(queryDscr);
    }
  }
  free(args);
  free(progname());
  return 0;
}


void runAnalysis(int queryDscr, Args *args){
  Sequence *query, *sbjct;
  Sequence *seq;
  int *sl;
  double p;
  int sbjctDscr;

  sbjctDscr = open(args->s,0);
  if(sbjctDscr < 0)
    eprintf("ERROR: could not open sbjct file %s\n",args->s);
  query = readFasta(queryDscr);
  sbjct = readFasta(sbjctDscr);
  close(sbjctDscr);
  prepareSeq(query);
  prepareSeq(sbjct);
  seq = catSeq(query,sbjct);
  seq->sbjctGc = gcContent(sbjct);
  seq->queryGc = gcContent(query);
  query = freeSequence(query);
  sbjct = freeSequence(sbjct);
  sl = getLcpTreeShulens(args, seq);
  if(args->w){
    windowAnalysis(sl,seq->numQueryNuc/2,args->w,args->S);
    return;
  }
  if(args->u)
    printShulens(sl, seq->numQueryNuc/2);
  else{
    p = pim(args, seq, sl);
    printf("pi_m: %8.4e\n",p);
  }
  freeSequence(seq);
}

void printShulens(int *sl, int n){
  int i;
  for(i=0;i<n;i++){
    printf("%d\t%d\n",i+1,sl[i]);
  }
}

/* windowAnalysis: sliding window analysis of shulens */
void windowAnalysis(int *sl, int n, int winLen, int stepLen){
  int i, j, s;
  double pos;
  double t;

  s = 0;
  for(i=0;i<winLen;i++)
    s += sl[i];
  pos = (double)i - winLen/2.;
  t = (double)winLen/(double)s;
  printf("%.1f\t%g\n",pos,t);
  
  for(i=winLen;i<n;i+=stepLen){
    for(j=0;j<stepLen;j++){
      s += sl[i+j];
      s -= sl[i+j-winLen];
    }
    pos = (double)i - winLen/2.;
    t = (double)winLen/(double)s;
    printf("%.1f\t%g\n",pos,t);
  }
}


double pim(Args *args, Sequence *seq, int *sl){
  int i;
  Result *res;
  double avgShulen;

  res = countShulens(sl,seq->numQueryNuc);  

  avgShulen = 0;
  for(i=0;i<res->n;i++)
    avgShulen += res->c[i] * (i);
  
  return seq->numSbjctNuc/avgShulen;
}

Result *countShulens(int *sl, int n){
  int max = -1;
  Result *res;
  int i, *c;

  /* find maximum */
  for(i=0;i<n;i++)
    if(max < sl[i])
      max = sl[i];
  c = (int *)emalloc((max+1)*sizeof(int));
  for(i=0;i<=max;i++)
    c[i] = 0;
  for(i=0;i<n;i++)
    c[sl[i]]++;
  res = (Result *)emalloc(sizeof(Result));
  res->n = max+1;
  res->c = c;
  return res;
}
