/***** shulen.c **************************************************************
 * Description: Compute the lengths of shortest unique substrings.
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Wed Dec  6 16:12:32 2006.
 * 
 * This file is part of kr.
 *
 * kr is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * kr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with kr; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA.
 *****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "sequenceData.h"
#include "eprintf.h"
#include "interface.h"
#include "shulen.h"
#include "DeepShallow/ds_ssort.h"
#include "DeepShallow/bwt_aux.h"
#include "DeepShallow/lcp_aux.h"

int *getSuffixArray(Sequence *seq){
  long overshoot;
  int *sa;
  unsigned char *textu;

  /* init ds suffix sort routine (cf. DeepShallow/testlcp.c) */
  overshoot = init_ds_ssort(500,2000);
  if(overshoot==0)
    eprintf("ERROR [ir]: ds initialization failed.\n");
  sa = (int *)emalloc((seq->len+1)*sizeof(int));
  seq->seq = (char *)erealloc(seq->seq,(seq->len+overshoot)*sizeof(unsigned char));
  textu = (unsigned char *)seq->seq;
  ds_ssort(textu, sa+1, seq->len);
  return sa;
}

int *getLcp(Sequence *seq, int *sa){
  int occ[ALPHA_SIZE];
  unsigned char *textu;
  long i;
  
  textu = (unsigned char *)seq->seq;
  for(i=0;i<ALPHA_SIZE;i++)
    occ[i] = 0;
  for(i=0;i<seq->len;i++)
    occ[textu[i]]++;
  
  return _lcp_sa2lcp_9n(textu,seq->len,sa,occ);
}

/* getShulensWithoutSentinel: return array of true shustring lengths */
int *getShulensWithoutSentinel(Args *args, Sequence *seq){
  long i, j, min; 
  int *sl;
  int *sa, *lcp;

  sa = getSuffixArray(seq);
  lcp = getLcp(seq,sa);

  sl = getRawShulens(args, seq, sa, lcp);
  free(sa);
  free(lcp);
  /* filter out shulens that cross string borders */
  min = 0;
  for(i=0;i<seq->numSeq;i++){
    if(i)
      min = seq->borders[i-1];
    for(j=seq->borders[i]-1;j>min;j--)
      if(j + sl[j] > seq->borders[i])
	sl[j] = -1;
      else
	break;
  }
  return sl;
}

int *getRawShulens(Args *args, Sequence *seq, int *sa, int *lcp){
  long i, arrayLen;
  int *sl;
  
  arrayLen = seq->len;
  sl = (int *)emalloc(seq->len*sizeof(int));
  for(i=2;i<arrayLen;i++)
    if(sa[i]<seq->len){
      if(lcp[i+1] < lcp[i])
	sl[sa[i]] = lcp[i] + 1;
      else
	sl[sa[i]] = lcp[i+1] + 1;
    }
  if(sa[i] < seq->len)
    sl[sa[i]] = lcp[i] + 1;
  if(sa[1] < seq->len)
    sl[sa[1]] = lcp[2] + 1;
  return sl;
}

/* getShulensWithSentinel: return shulens as if each sequence was terminated by a unique character. */
int *getShulensWithSentinel(Args *args, Sequence *seq){
  long i, j, min;
  int *sl, *sa, *lcp;

  sa = getSuffixArray(seq);
  lcp = getLcp(seq,sa);
  sl = getRawShulens(args, seq, sa, lcp);
  free(sa);
  free(lcp);
  /* prune shulens that cross string borders */

  min = 0;
  for(i=0;i<seq->numSeq;i++){
    if(i)
      min = seq->borders[i-1];
    for(j=seq->borders[i]-1;j>min;j--)
      if(j + sl[j] > seq->borders[i])
	sl[j] = seq->borders[i] - j + 1;
      else
	break;
  }

  return sl;
}

void printSuffixArray(Args *args, int *sa, int *lcpTab, int n, char *seq){
  int i, j;

  for(i=0;i<n;i++){
    printf("%d\t%d\t%d\t",i,sa[i],lcpTab[i]);
    for(j=sa[i];j<n;j++)
      printf("%c",seq[j]);
    printf("\n");
  }

}
